from cryptography.fernet import Fernet
from django.conf import settings


class EncryptedFileMixin:
    
    def __init__(self):
        if not hasattr(settings, 'ENCRYPTION_KEY'):
            raise ValueError("ENCRYPTION_KEY must be set in your environment.")
        self.cipher = Fernet(settings.ENCRYPTION_KEY)

    def encrypted(self, content):
        """
        Criptografa o conteúdo.
        """

        if isinstance(content, str):
            content = content.encode()  # Garante que o conteúdo seja convertido para bytes

        elif not isinstance(content, bytes):
            raise ValueError("O conteúdo precisa ser uma string ou bytes para criptografia.")
        
        return self.cipher.encrypt(content)

    def decrypted(self, content):
        """
        Descriptografa o conteúdo.
        """
        
        try:
            return self.cipher.decrypt(content)
        except Exception as e:
            raise ValueError(f"Erro ao descriptografar: {str(e)}")  # Exceção mais clara



from django.contrib.auth.decorators import login_required
from django.contrib.auth.decorators import permission_required

from django.apps import apps
from django.http import HttpResponse
from django.db.models.fields.files import FieldFile
from mimetypes import guess_type
from django.core.cache import cache

from encrypted_fields.encrypted_fields import *
from encrypted_fields.encrypted_files import *


# Create your views here.
def serve_decrypted_file(request, app_name, model_name, field_name, pk):
    """
    View para descriptografar e retornar o arquivo (imagem ou qualquer outro) com cache de 5 minutos.
    """
    try:
        # Tenta recuperar o arquivo do cache
        cache_key = f"{app_name}_{model_name}_{field_name}_{pk}"
        cached_file = cache.get(cache_key)

        if cached_file:
            return cached_file  # Retorna o arquivo cacheado se disponível

        # Valida se o app existe
        if app_name not in apps.app_configs:
            return HttpResponse("App não encontrado.", status=404)

        # Obtém o modelo dinamicamente a partir do nome
        model = apps.get_model(app_name, model_name)
        if not model:
            return HttpResponse("Modelo não encontrado.", status=404)
        
        # Busca o objeto pelo ID (pk)
        obj = model.objects.filter(pk=pk).first()
        if not obj:
            return HttpResponse("Objeto não encontrado.", status=404)
        
        # Obtém o campo do objeto dinamicamente
        file_field = getattr(obj, field_name, None)
        if not file_field:
            return HttpResponse("Campo não encontrado.", status=404)
        
        # Verifica se o campo é do tipo EncryptedFileField ou EncryptedImageField
        if isinstance(file_field, (EncryptedFileField, EncryptedImageField)):
            # Chama a descriptografia via from_db_value
            decrypted_file = file_field.from_db_value(file_field.name, None, None)
            if not decrypted_file:
                return HttpResponse("Erro ao descriptografar o arquivo.", status=500)
        elif isinstance(file_field, (FieldFile)):
            # Se for um FileField ou ImageField normal, abre o arquivo associado
            decrypted_file = file_field.file
        else:
            return HttpResponse("Campo não é do tipo FieldFile.", status=400)

        # Determina o tipo MIME do arquivo
        mime_type, _ = guess_type(decrypted_file.name)
        
        # Cria a resposta com o arquivo
        response = HttpResponse(decrypted_file.read(), content_type=mime_type)
        response['Content-Disposition'] = f'inline; filename="{decrypted_file.name}"'

        # Armazena a resposta no cache por 5 minutos
        cache.set(cache_key, response, timeout=300)

        return response
        
    except Exception as e:
        return HttpResponse(f"Erro: {e}", status=500)



from django.db import models
from datetime import datetime
from .encrypted_field_mixin import EncryptedFieldMixin


class EncryptedIntegerField(EncryptedFieldMixin, models.IntegerField):
    """
    Campo criptografado para valores inteiros.
    """
    def cast_value(self, value):
        return int(value)


class EncryptedFloatField(EncryptedFieldMixin, models.FloatField):
    """
    Campo criptografado para valores flutuantes (float).
    """
    def cast_value(self, value):
        return float(value)


class EncryptedBooleanField(EncryptedFieldMixin, models.BooleanField):
    """
    Campo criptografado para valores booleanos.
    """
    def cast_value(self, value):
        return value.lower() == 'true'


class EncryptedCharField(EncryptedFieldMixin, models.CharField):
    """
    Campo criptografado para strings.
    """
    def cast_value(self, value):
        return value  # Strings não precisam de conversão adicional


class EncryptedTextField(EncryptedFieldMixin, models.TextField):
    """
    Campo criptografado para strings.
    """
    def cast_value(self, value):
        return value  # Strings não precisam de conversão adicional


class EncryptedDateField(EncryptedFieldMixin, models.DateField):
    """
    Campo criptografado para datas.
    """
    def cast_value(self, value):
        if isinstance(value, str):
            # Supondo que o formato da data armazenada seja 'YYYY-MM-DD'
            return datetime.strptime(value, '%Y-%m-%d').date()
        return value  # Se já for uma instância de `date`, não é necessário converter


class EncryptedDateTimeField(EncryptedFieldMixin, models.DateTimeField):
    """
    Campo criptografado para valores de data e hora.
    """
    def cast_value(self, value):
        if isinstance(value, str):
            # Supondo que o formato seja 'YYYY-MM-DD HH:MM:SS'
            return datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
        return value


class EncryptedTimeField(EncryptedFieldMixin, models.TimeField):
    """
    Campo criptografado para valores de hora.
    """
    def cast_value(self, value):
        if isinstance(value, str):
            # Supondo que o formato seja 'HH:MM:SS'
            return datetime.strptime(value, '%H:%M:%S').time()
        return value


class EncryptedDecimalField(EncryptedFieldMixin, models.DecimalField):
    """
    Campo criptografado para valores decimais.
    """
    def cast_value(self, value):
        from decimal import Decimal
        return Decimal(value)


class EncryptedEmailField(EncryptedFieldMixin, models.EmailField):
    """
    Campo criptografado para endereços de e-mail.
    """
    def cast_value(self, value):
        return value  # Strings de e-mail não precisam de conversão adicional


class EncryptedURLField(EncryptedFieldMixin, models.URLField):
    """
    Campo criptografado para URLs.
    """
    def cast_value(self, value):
        return value  # Strings de URL não precisam de conversão adicional


class EncryptedUUIDField(EncryptedFieldMixin, models.UUIDField):
    """
    Campo criptografado para valores UUID.
    """
    def cast_value(self, value):
        import uuid
        if isinstance(value, str):
            return uuid.UUID(value)
        return value


class EncryptedJSONField(EncryptedFieldMixin, models.JSONField):
    """
    Campo criptografado para valores JSON.
    """
    def cast_value(self, value):
        import json
        if isinstance(value, str):
            return json.loads(value)
        return value  # Assume que já é um dicionário ou lista



from django.db import models
from django.core.files.base import ContentFile
from io import BytesIO
from .encrypted_file_mixin import EncryptedFileMixin
from django.core.files.storage import default_storage
from PIL import Image


class EncryptedFileField(models.FileField):
    """
    Um campo de arquivo criptografado.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cryptographer = EncryptedFileMixin()  # Instanciando o EncryptedFileMixin para usá-lo

    def pre_save(self, model_instance, add):
        """
        Antes de salvar o arquivo, criptografa o conteúdo.
        """
        file = getattr(model_instance, self.attname)
        
        if file:
            if file.size == 0:  # Verifica se o arquivo está vazio
                model_instance.__dict__[self.attname] = None  # Define como None se estiver vazio
            else:
                encrypted_file = self._encrypt_file(file)  # Criptografa o arquivo
                model_instance.__dict__[self.attname] = encrypted_file
        else:
            model_instance.__dict__[self.attname] = None  # Define como None se não houver arquivo
        
        return super().pre_save(model_instance, add)

    def _encrypt_file(self, file):
        """
        Função para criptografar o arquivo antes de ser salvo.
        """
        content = file.read()  # Lê o conteúdo do arquivo
        encrypted_content = self.cryptographer.encrypted(content)  # Chama o método encrypted() da instância

        if isinstance(encrypted_content, bytes):
            encrypted_file = BytesIO(encrypted_content)
            encrypted_file.name = file.name  # Mantém o nome original do arquivo
            encrypted_file.size = len(encrypted_content)  # Ajusta o tamanho do arquivo
            encrypted_file.seek(0)  # Posiciona o ponteiro do arquivo no início
            return ContentFile(encrypted_file.read(), name=file.name)  # Retorna o arquivo criptografado como ContentFile
        else:
            raise ValueError("A criptografia falhou, o conteúdo não é um tipo de dado esperado.")

    def from_db_value(self, value, expression, connection, context=None):
        """
        Descriptografa o arquivo ao ser acessado do banco de dados.
        Se o valor for vazio ou None, retorna None.
        """
        if value is None or value == b'' or value == '':  # Verifica se o valor é vazio ou None
            return None

        file_path = value  # Supondo que o valor seja uma string representando o caminho do arquivo
        
        if default_storage.exists(file_path):
            # Lê o conteúdo do arquivo como bytes
            with default_storage.open(file_path, 'rb') as encrypted_file:  # Abre o arquivo para leitura binária
                return self._decrypt_file(encrypted_file)  # Passa os bytes para descriptografia
        else:
            raise ValueError(f"O arquivo no caminho {file_path} não existe.")

    def _decrypt_file(self, encrypted_file):
        """
        Função para descriptografar o conteúdo criptografado.
        """

        encrypted_content = encrypted_file.read()  # Lê todo o conteúdo como bytes
        decrypted_content = self.cryptographer.decrypted(encrypted_content)

        if isinstance(decrypted_content, bytes):
            # Cria um arquivo em memória com o conteúdo descriptografado
            decrypted_file = BytesIO(decrypted_content)
            decrypted_file.name = encrypted_file.name  # Mantém o nome original do arquivo
            decrypted_file.size = len(decrypted_content)  # Ajusta o tamanho do arquivo
            decrypted_file.seek(0)  # Posiciona o ponteiro do arquivo no início
            # Retorna o arquivo descriptografado como ContentFile
            return ContentFile(decrypted_file.read(), name=encrypted_file.name)
        else:
            raise ValueError("A decriptação falhou, o conteúdo não é um tipo de dado esperado.")


class EncryptedImageField(models.ImageField):
    """
    Um campo de imagem criptografado.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cryptographer = EncryptedFileMixin()  # Instanciando o EncryptedFileMixin para usá-lo

    def pre_save(self, model_instance, add):
        """
        Antes de salvar a imagem, criptografa o conteúdo.
        """
        image = getattr(model_instance, self.attname)

        if image:
            if image.size == 0:  # Verifica se o arquivo está vazio
                model_instance.__dict__[self.attname] = None  # Define como None se estiver vazio
            else:
                encrypted_image = self._encrypt_image(image)  # Criptografa a imagem
                model_instance.__dict__[self.attname] = encrypted_image
        else:
            model_instance.__dict__[self.attname] = None  # Define como None se não houver arquivo

        return super().pre_save(model_instance, add)

    def _encrypt_image(self, image):
        """
        Função para criptografar a imagem antes de ser salva.
        """
        content = image.read()  # Lê o conteúdo da imagem
        encrypted_content = self.cryptographer.encrypted(content)  # Criptografa o conteúdo

        if isinstance(encrypted_content, bytes):
            encrypted_image = BytesIO(encrypted_content)
            encrypted_image.name = image.name  # Mantém o nome original do arquivo
            encrypted_image.size = len(encrypted_content)  # Ajusta o tamanho do arquivo
            encrypted_image.seek(0)  # Posiciona o ponteiro no início
            return ContentFile(encrypted_image.read(), name=image.name)  # Retorna a imagem criptografada como ContentFile
        else:
            raise ValueError("A criptografia falhou, o conteúdo não é um tipo de dado esperado.")

    def from_db_value(self, value, expression, connection, context=None):
        """
        Descriptografa a imagem ao ser acessada do banco de dados.
        Se o valor for vazio ou None, retorna None.
        """
        if value is None or value == b'' or value == '':  # Verifica se o valor é vazio ou None
            return None

        file_path = value  # Supondo que o valor seja uma string representando o caminho do arquivo

        if default_storage.exists(file_path):
            # Lê o conteúdo do arquivo como bytes
            with default_storage.open(file_path, 'rb') as encrypted_file:  # Abre o arquivo para leitura binária
                return self._decrypt_image(encrypted_file)  # Passa os bytes para descriptografia
        else:
            raise ValueError(f"O arquivo no caminho {file_path} não existe.")

    def _decrypt_image(self, encrypted_file):
        """
        Função para descriptografar o conteúdo criptografado da imagem.
        """
        encrypted_content = encrypted_file.read()  # Lê todo o conteúdo como bytes
        decrypted_content = self.cryptographer.decrypted(encrypted_content)  # Descriptografa o conteúdo

        if isinstance(decrypted_content, bytes):
            # Cria uma imagem em memória com o conteúdo descriptografado
            decrypted_image = BytesIO(decrypted_content)
            decrypted_image.name = encrypted_file.name  # Mantém o nome original do arquivo
            decrypted_image.seek(0)  # Posiciona o ponteiro no início

            # Valida se o conteúdo é uma imagem válida
            try:
                Image.open(decrypted_image).verify()
                decrypted_image.seek(0)  # Reposiciona o ponteiro após a verificação
            except Exception as e:
                raise ValueError(f"A decriptação falhou, o conteúdo não é uma imagem válida: {e}")

            # Retorna a imagem descriptografada como ContentFile
            return ContentFile(decrypted_image.read(), name=encrypted_file.name)
        else:
            raise ValueError("A decriptação falhou, o conteúdo não é um tipo de dado esperado.")



from django.urls import path
from .views import *
from django.conf import settings

app_name = 'serve_files'
if not hasattr(settings, 'SERVE_DECRYPTED_FILE_URL_BASE'):
    raise ValueError("SERVE_DECRYPTED_FILE_URL_BASE must be set in your environment.")
base_url_serve_decrypted_file = settings.SERVE_DECRYPTED_FILE_URL_BASE

urlpatterns = [
  path(f'{base_url_serve_decrypted_file}/<str:app_name>/<str:model_name>/<str:field_name>/<int:pk>/', serve_decrypted_file, name='serve_decrypted_file'),
]



from setuptools import setup, find_packages

setup(
    name="django-encrypted-fields-and-files",
    version="0.1.1",
    packages=find_packages(),
    install_requires=["Django>=4.2", "cryptography>=41.0.5", "pillow>=10.0.0"],
    author="Denky",
    author_email="contato@denky.dev.br",
    description="Campos e storages criptografados para o Django.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/D3NKYT0/django-encrypted-fields",
    project_urls={
        "Source": "https://github.com/D3NKYT0/django-encrypted-fields",
        "Homepage": "https://denky.dev.br/django-encrypted-fields",
        "Documentation": "https://github.com/D3NKYT0/django-encrypted-fields/wiki",
        "Issue Tracker": "https://github.com/D3NKYT0/django-encrypted-fields/issues",
    },
    classifiers=[
        "Framework :: Django",
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.10",
)
